@@ -1,8 +1,11 @@
+from __future__ import print_function
+
 import gzip
 import random
 import time
-import cPickle
+import pickle
 
+from neat.config import Config
 from neat.genome import Genome, FFGenome
 from neat.genes import NodeGene, ConnectionGene
 from neat.species import Species
@@ -16,7 +19,13 @@ class Population(object):
     def __init__(self, config, checkpoint_file=None, initial_population=None,
                  node_gene_type=NodeGene, conn_gene_type=ConnectionGene,
                  diversity_type=AgedFitnessSharing):
+
+        # If config is not a Config object, assume it is a path to the config file.
+        if not isinstance(config, Config):
+            config = Config(config)
+
         self.config = config
+
         self.population = None
         self.node_gene_type = node_gene_type
         self.conn_gene_type = conn_gene_type
@@ -49,24 +58,24 @@ class Population(object):
         # TODO: Wouldn't it just be better to create a class method to load and return the stored Population
         # object as-is?  I don't know if there are hidden side effects to directly replacing __dict__.
         with gzip.open(checkpoint) as f:
-            print 'Resuming from a previous point: %s' % checkpoint
+            print('Resuming from a previous point: %s' % checkpoint)
             # when unpickling __init__ is not called again
-            previous_pop = cPickle.load(f)
+            previous_pop = pickle.load(f)
             self.__dict__ = previous_pop.__dict__
 
-            print 'Loading random state'
-            random.setstate(cPickle.load(f))
+            print('Loading random state')
+            random.setstate(pickle.load(f))
 
     def __create_checkpoint(self, report):
         """ Saves the current simulation state. """
         if report:
-            print 'Creating checkpoint file at generation: %d' % self.generation
+            print('Creating checkpoint file at generation: %d' % self.generation)
 
         with gzip.open('checkpoint_' + str(self.generation), 'w', compresslevel=5) as f:
             # Write the entire population state.
-            cPickle.dump(self, f, protocol=cPickle.HIGHEST_PROTOCOL)
+            pickle.dump(self, f, protocol=pickle.HIGHEST_PROTOCOL)
             # Remember the current random number state.
-            cPickle.dump(random.getstate(), f, protocol=2)
+            pickle.dump(random.getstate(), f, protocol=2)
 
     def __create_population(self):
         if self.config.feedforward:
@@ -75,12 +84,18 @@ class Population(object):
             genotypes = Genome
 
         self.population = []
+        # TODO: Add FS-NEAT support, which creates an empty connection set, and then performs a
+        # single add connection mutation.
+        # This would give three initialization methods:
+        # 1. Fully connected (each input connected to all outputs)
+        # 2. "minimally" connected, which isn't really minimal (one random input to each output)
+        # 3. FS-NEAT connected (one random connection)
         if self.config.fully_connected:
-            for i in xrange(self.config.pop_size):
+            for i in range(self.config.pop_size):
                 g = genotypes.create_fully_connected(self.config, self.node_gene_type, self.conn_gene_type)
                 self.population.append(g)
         else:
-            for i in xrange(self.config.pop_size):
+            for i in range(self.config.pop_size):
                 g = genotypes.create_minimally_connected(self.config, self.node_gene_type, self.conn_gene_type)
                 self.population.append(g)
 
@@ -94,7 +109,7 @@ class Population(object):
         return s
 
     def __speciate(self, report):
-        """ Group chromosomes into species by similarity """
+        """ Group genomes into species by similarity """
         # Speciate the population
         for individual in self.population:
             found = False
@@ -104,17 +119,17 @@ class Population(object):
                     found = True
                     break
 
-            if not found:  # create a new species for this lone chromosome
+            if not found:  # create a new species for this lone genome
                 self.__species.append(Species(individual))
 
         # python technical note:
         # we need a "working copy" list when removing elements while looping
         # otherwise we might end up having sync issues
         for s in self.__species[:]:
-            # this happens when no chromosomes are compatible with the species
-            if len(s.members) == 0:
+            # this happens when no genomes are compatible with the species
+            if not s.members:
                 if report:
-                    print "Removing species %d for being empty" % s.ID
+                    print("Removing species %d for being empty" % s.ID)
                 # remove empty species
                 self.__species.remove(s)
 
@@ -136,17 +151,18 @@ class Population(object):
 
     def __log_species(self):
         """ Logging species data for visualizing speciation """
-        higher = max([s.ID for s in self.__species])
         temp = []
-        for i in xrange(1, higher + 1):
-            found_species = False
-            for s in self.__species:
-                if i == s.ID:
-                    temp.append(len(s.members))
-                    found_species = True
-                    break
-            if not found_species:
-                temp.append(0)
+        if self.__species:
+            higher = max([s.ID for s in self.__species])
+            for i in range(1, higher + 1):
+                found_species = False
+                for s in self.__species:
+                    if i == s.ID:
+                        temp.append(len(s.members))
+                        found_species = True
+                        break
+                if not found_species:
+                    temp.append(0)
         self.species_log.append(temp)
 
     def epoch(self, fitness_function, n, report=True, save_best=False, checkpoint_interval=10,
@@ -155,25 +171,25 @@ class Population(object):
 
             Keyword arguments:
             report -- show stats at each epoch (default True)
-            save_best -- save the best chromosome from each epoch (default False)
+            save_best -- save the best genome from each epoch (default False)
             checkpoint_interval -- time in minutes between saving checkpoints (default 10 minutes)
             checkpoint_generation -- time in generations between saving checkpoints
                 (default None -- option disabled)
         """
         t0 = time.time()  # for saving checkpoints
 
-        for g in xrange(n):
+        for g in range(n):
             self.generation += 1
 
             if report:
-                print '\n ****** Running generation %d ****** \n' % self.generation
+                print('\n ****** Running generation %d ****** \n' % self.generation)
 
             # Evaluate individuals
             fitness_function(self.population)
             # Speciates the population
             self.__speciate(report)
 
-            # Current generation's best chromosome
+            # Current generation's best genome
             self.most_fit_genomes.append(max(self.population))
             # Current population's average fitness
             self.avg_fitness_scores.append(mean([c.fitness for c in self.population]))
@@ -181,25 +197,25 @@ class Population(object):
             # Print some statistics
             best = self.most_fit_genomes[-1]
 
-            # saves the best chromo from the current generation
+            # saves the best genome from the current generation
             if save_best:
-                f = open('best_chromo_' + str(self.generation), 'w')
-                cPickle.dump(best, f)
+                f = open('best_genome_' + str(self.generation), 'w')
+                pickle.dump(best, f)
                 f.close()
 
             # Stops the simulation
             if best.fitness > self.config.max_fitness_threshold:
                 if report:
-                    print '\nBest individual in epoch %s meets fitness threshold - complexity: %s' % (
-                        self.generation, best.size())
+                    print('\nBest individual in epoch %s meets fitness threshold - complexity: %s' % (
+                        self.generation, best.size()))
                 break
 
-            # Remove stagnated species and its members (except if it has the best chromosome)
+            # Remove stagnated species and its members (except if it has the best genome)
             for s in self.__species[:]:
                 if s.no_improvement_age > self.config.max_stagnation:
                     if report:
-                        print "\n   Species %2d (with %2d individuals) is stagnated: removing it" \
-                              % (s.ID, len(s.members))
+                        print("\n   Species %2d (with %2d individuals) is stagnated: removing it" \
+                              % (s.ID, len(s.members)))
                     # removing species
                     self.__species.remove(s)
                     # removing all the species' members
@@ -216,7 +232,7 @@ class Population(object):
                 # This rarely happens
                 if s.spawn_amount == 0:
                     if report:
-                        print '   Species %2d age %2s removed: produced no offspring' % (s.ID, s.age)
+                        print('   Species %2d age %2s removed: produced no offspring' % (s.ID, s.age))
                     for c in self.population[:]:
                         if c.species_id == s.ID:
                             self.population.remove(c)
@@ -227,60 +243,67 @@ class Population(object):
             self.__log_species()
 
             if report:
-                std_dev = stdev([c.fitness for c in self.population])
-                print 'Population\'s average fitness: %3.5f stdev: %3.5f' % (self.avg_fitness_scores[-1], std_dev)
-                print 'Best fitness: %2.12s - size: %s - species %s - id %s' \
-                      % (best.fitness, best.size(), best.species_id, best.ID)
-                print 'Species length: %d totaling %d individuals' \
-                      % (len(self.__species), sum([len(s.members) for s in self.__species]))
-                print 'Species ID       : %s' % [s.ID for s in self.__species]
-                print 'Each species size: %s' % [len(s.members) for s in self.__species]
-                print 'Amount to spawn  : %s' % [s.spawn_amount for s in self.__species]
-                print 'Species age      : %s' % [s.age for s in self.__species]
-                print 'Species no improv: %s' % [s.no_improvement_age for s in self.__species]
+                if self.population:
+                    std_dev = stdev([c.fitness for c in self.population])
+                    print('Population\'s average fitness: %3.5f stdev: %3.5f' % (self.avg_fitness_scores[-1], std_dev))
+                    print('Best fitness: %2.12s - size: %s - species %s - id %s' \
+                          % (best.fitness, best.size(), best.species_id, best.ID))
+                    print('Species length: %d totaling %d individuals' \
+                          % (len(self.__species), sum([len(s.members) for s in self.__species])))
+                    print('Species ID       : %s' % [s.ID for s in self.__species])
+                    print('Each species size: %s' % [len(s.members) for s in self.__species])
+                    print('Amount to spawn  : %s' % [s.spawn_amount for s in self.__species])
+                    print('Species age      : %s' % [s.age for s in self.__species])
+                    print('Species no improv: %s' % [s.no_improvement_age for s in self.__species])
+                else:
+                    print('All species extinct.')
 
             # -------------------------- Producing new offspring -------------------------- #
             new_population = []  # next generation's population
 
-            # Spawning new population
-            for s in self.__species:
-                new_population.extend(s.reproduce(self.config))
+            # If no species are left, create a new population from scratch, otherwise top off
+            # population by reproducing existing species.
+            if self.__species:
+                for s in self.__species:
+                    new_population.extend(s.reproduce(self.config))
 
-            # Controls under or overflow  #
-            fill = self.config.pop_size - len(new_population)
-            if fill < 0:  # overflow
-                if report:
-                    print '   Removing %d excess individual(s) from the new population' % -fill
-                # TODO: This is dangerous! I can't remove a species' representative!
-                new_population = new_population[:fill]  # Removing the last added members
+                # Controls under or overflow  #
+                fill = self.config.pop_size - len(new_population)
+                if fill < 0:  # overflow
+                    if report:
+                        print('   Removing %d excess individual(s) from the new population' % -fill)
+                    # TODO: This is dangerous! I can't remove a species' representative!
+                    new_population = new_population[:fill]  # Removing the last added members
 
-            if fill > 0:  # underflow
-                if report:
-                    print '   Producing %d more individual(s) to fill up the new population' % fill
+                if fill > 0:  # underflow
+                    if report:
+                        print('   Producing %d more individual(s) to fill up the new population' % fill)
 
-                # TODO: what about producing new individuals instead of reproducing?
-                # increasing diversity from time to time might help
-                while fill > 0:
-                    # Selects a random chromosome from population
-                    parent1 = random.choice(self.population)
-                    # Search for a mate within the same species
-                    found = False
-                    for c in self.population:
-                        # what if c is parent1 itself?
-                        if c.species_id == parent1.species_id:
-                            child = parent1.crossover(c)
-                            new_population.append(child.mutate())
-                            found = True
-                            break
-                    if not found:
-                        # If no mate was found, just mutate it
-                        new_population.append(parent1.mutate())
-                    # new_population.append(chromosome.FFGenome.create_fully_connected())
-                    fill -= 1
+                    # TODO: what about producing new individuals instead of reproducing?
+                    # increasing diversity from time to time might help
+                    while fill > 0:
+                        # Selects a random genome from population
+                        parent1 = random.choice(self.population)
+                        # Search for a mate within the same species
+                        found = False
+                        for c in self.population:
+                            # what if c is parent1 itself?
+                            if c.species_id == parent1.species_id:
+                                child = parent1.crossover(c)
+                                new_population.append(child.mutate())
+                                found = True
+                                break
+                        if not found:
+                            # If no mate was found, just mutate it
+                            new_population.append(parent1.mutate())
+                        # new_population.append(genome.FFGenome.create_fully_connected())
+                        fill -= 1
 
-            assert self.config.pop_size == len(new_population), 'Different population sizes!'
-            # Updates current population
-            self.population = new_population[:]
+                assert self.config.pop_size == len(new_population), 'Different population sizes!'
+                # Updates current population
+                self.population = new_population[:]
+            else:
+                self.__create_population()
 
             if checkpoint_interval is not None and time.time() > t0 + 60 * checkpoint_interval:
                 self.__create_checkpoint(report)
